#!/usr/bin/env python

#
# This file is part of the `neko` Python module
#
# Copyright 2023
# Insitut Curie
#
# File author(s): Marco Ruscone (marco.ruscone@curie.fr)
#
# Distributed under the GPLv3 license
# See the file `LICENSE` or read a copy at
# https://www.gnu.org/licenses/gpl-3.0.txt
#

"""
Brief description of the project/module.
"""

__all__ = [
    '__version__',
    '__author__',
]

from ._metadata import __author__, __version__
